# Copyright 2023 Proyectos y Sistemas de Mantenimiento SL (eProsima).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
fastdds xml sub-command.

This verb validates the Fast DDS XML profiles files.

usage: fastdds xml [<xml-command>]

xml-commands:

    validate    validate XML profiles files using an XSD schema

positional arguments:
    command     xml-command to run

optional arguments:
    -h, --help  show this help message and exit
"""

import argparse
import os.path
from pathlib import Path

from xml_ci.validate import Validate


class XMLParser:
    """XML sub-commands parser."""

    def __init__(self, argv):
        """Parse the sub-command and dispatch to the appropriate handler.

        Shows usage if no sub-command is specified.

        Supported sub-commands:

            validate   validate XML profiles files using an XSD schema

        param argv list(str):
            list containing the arguments for the command
        """
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description="""
                Script for the validation of the XML files that define the
                profiles of the DDS entities in Fast DDS.
            """,
            usage='fastdds xml [<xml-command>]',
        )
        parser.add_argument(
            'validate',
            nargs='*',
            help='validate XML profiles files using an XSD schema'
        )
        parser.add_argument(
            '-d',
            '--debug',
            action='store_true',
            help='print debug info'
        )
        parser.add_argument(
            '-x',
            '--xsd-file',
            help='Fast DDS XSD schema for validation'
        )
        args = parser.parse_args(argv)

        try:
            if not os.path.isfile(args.xsd_file):
                print(f'The XSD schema does not exist: {args.xsd_file}. '
                      'Using default schema.')

                args.xsd_file = self.xsd_dir()
        # xsd schema argument not set, do not print error message
        except TypeError:
            args.xsd_file = self.xsd_dir()  # get default schema path

        if args.validate:
            if args.validate[0] == 'validate':
                args.validate.pop(0)
                if not Validate(args.xsd_file).run(args.validate):
                    exit(1)
            else:
                print(f'xml-command "{args.validate[0]}" is not valid')
        else:
            parser.print_help()
            exit(1)

    def xsd_dir(self):
        """
        Calculate the xsd directory.

        :return: The path to the platform specific the XSD directory

        """
        tool_path = Path(os.path.dirname(os.path.realpath(__file__)))
        # We assume the schema path is relative to our installation path
        return tool_path / '../../../share/fastRTPS_profiles.xsd'



# Copyright 2023 Proyectos y Sistemas de Mantenimiento SL (eProsima).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Sub-Command Validate implementation.

This sub-command validates the Fast DDS XML configuration files using an
XSD schema.
"""

import os
import sys

try:  # Try catch for new python dependency
    from xml import etree
except ImportError:
    print(
        'xml module not found. Try to install running "pip install xml"')
    sys.exit(1)

try:  # Try catch for new python dependency
    from xmlschema import XMLSchema, XMLSchemaValidationError
except ImportError:
    print(
        'xmlschema module not found. '
        'Try to install running "pip install xmlschema"')
    sys.exit(1)


class Validate:
    """Fast DDS XML configuration files validator."""

    def __init__(self, xsd_file):
        """Initialize Validate class."""
        self.xsd_file = xsd_file
        self.__visited_dirs = []

    @property
    def xsd_file(self):
        """XSD schema getter."""
        return self.__xsd_file

    @xsd_file.setter
    def xsd_file(self, xsd_file):
        """XSD schema setter."""
        if not os.path.isfile(xsd_file):
            print(f'The XSD schema does not exist: {xsd_file}')
            sys.exit(1)
        if isinstance(xsd_file, XMLSchema):
            self.__xsd_file = xsd_file
        else:
            try:
                self.__xsd_file = XMLSchema(xsd_file)
            except etree.ElementTree.ParseError as exception:
                print(f'cannot load XSD file: {xsd_file}')
                print(exception)
                sys.exit(1)

    def run(self, validate_list):
        """
        Run Fast DDS XML configuration files validation.

        :param validate_list: List with the directories and files to validate.
        :return: False if failed parsing, otherwise True.
        """
        valid = True
        self.__visited_dirs.clear()
        for element in validate_list:
            element = os.path.abspath(element)
            if os.path.isdir(element):
                valid = valid and self.__validate_xml_from_dir(element)
            elif os.path.isfile(element):
                if self.__is_xml(element):
                    valid = valid and self.__validate_xml_file(element)
                else:
                    print(f'The file is not an XML file: {element}')
                    valid = False
            else:
                print(f'{element} path not found')
                valid = False
        return valid

    def __validate_xml_from_dir(self, directory, recursive=True):
        """
        Validate XML files from a given directory.

        :param directory: The directory containing the XML files.
        :param recursive: True to do an iterative search for XML files
            through all the directories contained in the given directory.
        :return: False if failed parsing, otherwise True.
        """
        valid = True
        for root, _, files in os.walk(directory):
            print(f'Scanning directory {root}')
            if root not in self.__visited_dirs:
                self.__visited_dirs.append(root)
                for file in files:
                    if self.__is_xml(file):
                        valid = valid and self.__validate_xml_file(
                            os.path.join(root, file))
                if not recursive:
                    break
        return valid

    def __validate_xml_file(self, file):
        """
        Validate a single Fast DDS XML configuration file.

        :param file: The Fast DDS XML configuration file.
        :return: False if failed parsing, otherwise True.

        """
        try:
            xml_etree = etree.ElementTree.parse(file)
        except etree.ElementTree.ParseError as exception:
            print(f'cannot parse file: {file}')
            print(exception)
            sys.exit(1)

        try:
            self.xsd_file.validate(xml_etree)
            print(f'Valid XML file: {file}')
        except XMLSchemaValidationError as error:
            print(f'NOT valid XML file: {file}')
            print(error)
            return False
        return True

    def __is_xml(self, file):
        """
        Check if a file is an XML file.

        :param file: The input file
        :return: True if the file is an xml file, False if not.
        """
        return os.path.splitext(str(file))[-1] == '.xml'
